a {
  color: var(--yellow);
}
.btn-link {
  color: var(--light);
}
.content-div-right pre {
  font-family: Consolas, Menlo, Monaco, Lucida Console, Liberation Mono, DejaVu Sans Mono, Bitstream Vera Sans Mono, Courier New, monospace, serif;
  margin-bottom: 10px;
  overflow: auto;
  width: 100ch;
  padding: 5px;
  width: 650px!ie7;
  padding-bottom: 20px!ie7;
  max-height: 600px;
}
#console-div {
  overflow-y: scroll;
  max-height: 85vh;
  height: 85vh;
  color: white;
  background-color: black;
  padding: 0.5em;
}
.codeblock {
  background-color: var(--secondary) !important;
  color: white !important;
  display: block !important;
  overflow-x: auto !important;
  padding: 0.5em !important;

}
.running-job {
  color: green;
}
.stopped-job {
  color: red;
}
/* .dbc_dark code { */
  /* color: var(--light) !important; */
  /* background: var(--blue) !important; */
/* } */


/* .dash-dropdown { */
  /* color: #52575c; */
  /* background-color: #222; */
/* } */
/* .dash-dropdown .Select-control { */
  /* height: calc(1.5em + 0.75rem + 2px); */
  /* font-size: 0.9375rem; */
  /* font-weight: 400; */
  /* line-height: 1.5; */
  /* color: #fff; */
  /* background-color: #222; */
  /* background-clip: padding-box; */
  /* border: 1px solid #444; */
  /* border-radius: 0.25rem; */
  /* transition: border-color 0.15s ease-in-out, box-shadow 0.15s ease-in-out; */
/* } */

/* @media (prefers-reduced-motion: reduce) { */
  /* .dash-dropdown .Select-control { */
    /* transition: none; */
  /* } */
/* } */

/* .dash-dropdown .Select-control:hover { */
  /* box-shadow: none; */
/* } */

/* .dash-dropdown .is-open > .Select-control { */
  /* border-bottom-right-radius: 0; */
  /* border-bottom-left-radius: 0; */
  /* background-color: #222; */
/* } */
